-- this file is used with the SettingSaver plugin to dynamically access .plugindata file settings

-- This plugin helps provide real-time access to .plugindata files
-- The function SaveSetting and LoadSetting allow users asynchronous real-time access to .plugindata files (as little delay as possible due to the need to process requests through an external plugin)
-- Since the process is asynchronous, multiple calls to SaveSetting or LoadSetting can be queued
-- LoadSetting will either return an individual setting or if the requested setting is a sub-table, it will return the list of top level keys to that table
-- This system is safe to use with multiple simultaneous plugins since it internally implements the 0SSL command as a semaphore to limit access to one plugin at a time.

string.split=function(str,separator)
	local tmpArray={};
	local tmpElem;
	str=tostring(str);
	if string.find("^$()%.[]*+-?",separator,1,true)~=nil then separator="%"..separator end
	for tmpElem in string.gmatch(str,"[^"..separator.."]*") do
		table.insert(tmpArray,tmpElem);
	end
	local tmpIndex=1;
	while tmpIndex<=#tmpArray do
		if tmpIndex==#tmpArray and tmpArray[tmpIndex]=="" then
			table.remove(tmpArray,tmpIndex);
		else
			if tmpArray[tmpIndex]=="" and tmpArray[tmpIndex+1]~="" then
				table.remove(tmpArray,tmpIndex);
			else
				tmpIndex=tmpIndex+1;
			end
		end
	end
	return tmpArray;
end
string.encode=function(str)
	-- replace all non-alpha characters with a % escaped hex code
	local ret
	if str~=nil then
		if type(str)=="userdata" then
			-- calling routine tried to save a built-in C object such as a color, just suppress it
			ret=nil
		elseif type(str)=="number" then
			ret="#"..tostring(str)
		elseif type(str)=="boolean" then
			ret="$"..tostring(str)
		elseif type(str)=="string" then
			ret=""
			for i,v in ipairs({str:byte(1,-1)}) do
				if (v < 48) or (v>57 and v<65) or (v>90 and v<97) or (v > 125) then
					ret = ret .. "%" .. string.sub("00"..string.format("%x",tonumber(v)),-2)
				else
					ret = ret .. string.char(v)
				end
			end
		elseif type(str)=="table" then
			-- return list of keys
			ret="^"
			for k,v in pairs(str) do
				if type(k)=="number" then
					ret=ret.."|#"..tostring(k)
				else
					ret=ret.."|"..string.encode(k)
				end
			end
		else
			-- what do we do with values that aren't any of the expected types?
Turbine.Shell.WriteLine("unexpected type in string.encode:"..tostring(type(str)))
			ret=tostring(val)
		end
	end
	return ret
end
string.decode=function(str)
	local ret
	if str~=nil and type(str)=="string" then
		if string.sub(str,1,1)=="#" then
			ret=tonumber(string.sub(str,2))
		elseif string.sub(str,1,1)=="$" then
			ret=(string.sub(str,2)=="true")
		elseif string.sub(str,1,1)=="^" then
			ret=string.split(string.sub(str,3),"|")
			for k,v in pairs(ret) do
				if string.sub(v,1,1)=="#" then
					ret[k]=tonumber(string.sub(v,2))
				else
					ret[k]=string.decode(v)
				end
			end
		else
			ret=""
			while string.len(str)~=0 do
				local char=string.sub(str,1,1)
				if char~="%" then
					ret=ret..char
					str=string.sub(str,2,-1)
				else
					local char=string.char(tonumber("0x"..string.sub(str,2,3)))
					ret=ret..char
					str = string.sub(str,4,-1)
				end
			end
		end
	end
	return ret;
end

settingsaverwindow=Turbine.UI.Window()
settingsaverwindow.RequestQueue={}
settingsaverwindow.Cmds={}
settingsaverwindow.Update=function(sender,args)
	if settingsaverwindow.State=="init" then
		if not Turbine.Shell.IsCommand("0SSL") then
			-- use the 0SSL semaphore to prevent conclicts with other plugins using the same mechanism
			local newCmd=Turbine.ShellCommand()
			table.insert(settingsaverwindow.Cmds,newCmd)
			Turbine.Shell.AddCommand("0SSL",newCmd)
			settingsaverwindow.State="locked"
		end
	end
	-- use a seperate if clause so we can continue straight to processing as soon as we lock to avoid waiting another update cycle
	if settingsaverwindow.State=="locked" then
--Turbine.Shell.WriteLine("locked")
		-- set up parameters
		local newCmd=Turbine.ShellCommand()
		local tmpRequest={}
		if #settingsaverwindow.RequestQueue>0 then
			tmpRequest.Scope=settingsaverwindow.RequestQueue[1].Scope
			settingsaverwindow.CurrentFile=settingsaverwindow.RequestQueue[1].File
			if type(settingsaverwindow.RequestQueue[1].Branch)=="table" then
				settingsaverwindow.CurrentBranch=table.concat(settingsaverwindow.RequestQueue[1].Branch,"_")
			else
				settingsaverwindow.CurrentBranch=tostring(settingsaverwindow.RequestQueue[1].Branch)
			end
			tmpRequest.Recursive=settingsaverwindow.RequestQueue[1].Recursive
			tmpRequest.Debug=settingsaverwindow.RequestQueue[1].Debug
			tmpRequest.Key=settingsaverwindow.RequestQueue[1].Key
			settingsaverwindow.CurrentKey=tmpRequest.Key
			tmpRequest.Mode=settingsaverwindow.RequestQueue[1].Mode
			tmpRequest.Data=settingsaverwindow.RequestQueue[1].Data
			tmpRequest.Command=settingsaverwindow.RequestQueue[1].Command
			settingsaverwindow.Callback=settingsaverwindow.RequestQueue[1].Callback
			table.remove(settingsaverwindow.RequestQueue,1)
		else
			-- something went very wrong and we are out of synch
			-- abort processing and return
			settingsaverwindow.State="idle"
			settingsaverwindow:SetWantsUpdates(false)
			-- should probably try to cancel SettingSaver plugin too, but what if another plugin launched it? argh.
			return
		end
		if tmpRequest.Mode=="Delay" then
--Turbine.Shell.WriteLine("Mode=Delay #settingsaverwindow.RequestQueue:"..tostring(#settingsaverwindow.RequestQueue))
			if settingsaverwindow.Callback~=nil then
				settingsaverwindow.Callback()
			end
			if #settingsaverwindow.RequestQueue==0 then
				settingsaverwindow.State="idle"
				settingsaverwindow:SetWantsUpdates(false)
			else
				settingsaverwindow.State="init"
			end
			for k,v in pairs(settingsaverwindow.Cmds) do
				-- release the lock and any other commands
				Turbine.Shell.RemoveCommand(v)
			end
		elseif tmpRequest.Mode=="Exec" then
			table.insert(settingsaverwindow.Cmds,newCmd)
			Turbine.Shell.AddCommand("0SSS_"..tostring(tmpRequest.Scope),newCmd)
			newCmd=Turbine.ShellCommand()
			table.insert(settingsaverwindow.Cmds,newCmd)
			Turbine.Shell.AddCommand("0SSF_"..tostring(settingsaverwindow.CurrentFile),newCmd)
			newCmd=Turbine.ShellCommand()
			table.insert(settingsaverwindow.Cmds,newCmd)
			Turbine.Shell.AddCommand("0SSM_"..tostring(tmpRequest.Mode),newCmd)
			if tmpRequest.Debug then
				newCmd=Turbine.ShellCommand()
				table.insert(settingsaverwindow.Cmds,newCmd)
				Turbine.Shell.AddCommand("0SSD",newCmd)
			end
--Turbine.Shell.WriteLine("command:"..tostring(tmpRequest.Command))
			if tmpRequest.Command~=nil then
				local tmpCommand=string.encode(tmpRequest.Command)
				-- only encode existing values, others will be treated as nil on the receiving end (and values without a key will be ignored)
				local tmpPart=1
				while tmpCommand~=nil do
					newCmd=Turbine.ShellCommand()
					table.insert(settingsaverwindow.Cmds,newCmd)
--Turbine.Shell.WriteLine("AddCommand:".."0SSE"..tostring(tmpPart).."_"..string.sub(tmpCommand,1,115800))
					Turbine.Shell.AddCommand("0SSE"..tostring(tmpPart).."_"..string.sub(tmpCommand,1,115800),newCmd)
					tmpCommand=string.sub(tmpCommand,115801)
					if tmpCommand=="" then tmpCommand=nil end
					tmpPart=tmpPart+1
				end
			end
			-- launch SettingSaver plugin
			Turbine.PluginManager.LoadPlugin("SettingSaver");
			-- wait for resonse
			settingsaverwindow.State="waiting"
		else
			table.insert(settingsaverwindow.Cmds,newCmd)
			Turbine.Shell.AddCommand("0SSS_"..tostring(tmpRequest.Scope),newCmd)
			newCmd=Turbine.ShellCommand()
			table.insert(settingsaverwindow.Cmds,newCmd)
			Turbine.Shell.AddCommand("0SSF_"..tostring(settingsaverwindow.CurrentFile),newCmd)
			if tmpRequest.Recursive then
				newCmd=Turbine.ShellCommand()
				table.insert(settingsaverwindow.Cmds,newCmd)
				Turbine.Shell.AddCommand("0SST",newCmd)
			end
			if tmpRequest.Debug then
				newCmd=Turbine.ShellCommand()
				table.insert(settingsaverwindow.Cmds,newCmd)
				Turbine.Shell.AddCommand("0SSD",newCmd)
			end
			newCmd=Turbine.ShellCommand()
			table.insert(settingsaverwindow.Cmds,newCmd)
			Turbine.Shell.AddCommand("0SSM_"..tostring(tmpRequest.Mode),newCmd)
			local tmpKey
			local tmpBranch=settingsaverwindow.CurrentBranch
			newCmd=Turbine.ShellCommand()
			table.insert(settingsaverwindow.Cmds,newCmd)
			Turbine.Shell.AddCommand("0SSB_"..tmpBranch,newCmd)

			if tmpRequest.Key~=nil then
				if type(tmpRequest.Key)=="table" then
--Turbine.Shell.WriteLine("setting table of keys")
					for k,v in ipairs(tmpRequest.Key) do
						newCmd=Turbine.ShellCommand()
						table.insert(settingsaverwindow.Cmds,newCmd)
						Turbine.Shell.AddCommand("0SSK"..tostring(k).."_"..string.encode(v),newCmd)
						if tmpRequest.Mode=="Save" then
							tmpValue=nil
							if tmpRequest.Data~=nil and tmpRequest.Data[k]~=nil then
								local tmpData=string.encode(tmpRequest.Data[k])
								-- only encode existing values, others will be treated as nil on the receiving end (and values without a key will be ignored)
								local tmpPart=1
								while tmpData~=nil do
									newCmd=Turbine.ShellCommand()
									table.insert(settingsaverwindow.Cmds,newCmd)
									Turbine.Shell.AddCommand("0SSV"..tostring(k).."_"..tostring(tmpPart).."_"..string.sub(tmpData,1,115800),newCmd)
									tmpData=string.sub(tmpData,115801)
									if tmpData=="" then tmpData=nil end
									tmpPart=tmpPart+1
								end
							end
						end
					end
				else
--Turbine.Shell.WriteLine("setting key[0]:"..tostring(tmpRequest.Key))
					newCmd=Turbine.ShellCommand()
					table.insert(settingsaverwindow.Cmds,newCmd)
					Turbine.Shell.AddCommand("0SSK0_"..string.encode(tmpRequest.Key),newCmd)
					if tmpRequest.Mode=="Save" then
--Turbine.Shell.WriteLine("setting value[0]")
						tmpValue=nil
						if tmpRequest.Data~=nil then
							local tmpData=string.encode(tmpRequest.Data)
							-- only encode existing values, others will be treated as nil on the receiving end (and values without a key will be ignored)
							local tmpPart=1
							while tmpData~=nil do
								newCmd=Turbine.ShellCommand()
								table.insert(settingsaverwindow.Cmds,newCmd)
								Turbine.Shell.AddCommand("0SSV0_"..tostring(tmpPart).."_"..string.sub(tmpData,1,115800),newCmd)
								tmpData=string.sub(tmpData,115801)
								if tmpData=="" then tmpData=nil end
								tmpPart=tmpPart+1
							end
						end
					end
				end
			end

			-- launch SettingSaver plugin
			Turbine.PluginManager.LoadPlugin("SettingSaver");
			-- wait for resonse
			settingsaverwindow.State="waiting"
		end
--Turbine.Shell.WriteLine("waiting")
	end
	if settingsaverwindow.State=="waiting" then
		if Turbine.Shell.IsCommand("0SSC") then
--Turbine.Shell.WriteLine("complete")
			local retVal
			local tmpVal={}
			local tmpKey={}
			-- parse commands for 0SSRx values
			local cmds=Turbine.Shell.GetCommands();
			if cmds~=nil and type(cmds)=="table" then
				local tmp,tmpIndex
				table.sort(cmds,function(arg1,arg2)if arg1<arg2 then return(true) end end);
				for cmdIndex=0,#cmds do
--Turbine.Shell.WriteLine("returned cmd:"..tostring(cmds[cmdIndex]))
					if cmds[cmdIndex]~=nil then
						if cmdIndex>1 and cmds[cmdIndex]>="1" then
							break
						else
							tmpIndex,tmpPart,tmp=string.match(cmds[cmdIndex],"0SSR(.*)_(.*)_(.*)")
							if tmpIndex~=nil and tmpPart~=nil and tmp ~=nil then
								if tmpVal[tonumber(tmpIndex)]==nil then tmpVal[tonumber(tmpIndex)]={} end
								tmpVal[tonumber(tmpIndex)][tonumber(tmpPart)]=tmp
--Turbine.Shell.WriteLine("returned value:"..tostring(tmp))
							end
							tmpIndex,tmp=string.match(cmds[cmdIndex],"0SSZ(.*)_(.*)")
							if tmpIndex~=nil and tmp ~=nil then
								tmpKey[tonumber(tmpIndex)]=tmp
--Turbine.Shell.WriteLine("returned key:"..tostring(tmp))
							end
						end
					end
				end
			end
			-- reconstitute long strings
			for k,v in pairs(tmpVal) do
				local tmpRet=""
				for part,val in ipairs(v) do
					tmpRet=tmpRet..val
				end
				tmpVal[k]=tmpRet
			end
--Turbine.Shell.WriteLine("values:")
--Table.Dump(tmpVal)
--Turbine.Shell.WriteLine("keys:")
--Table.Dump(tmpKey)
			if tmpKey[0]~=nil then
				if tmpVal[0]==nil then
					-- return is a table of keys - not sure how useful this feature is now, but it is included for backward compatability with earlier versions of the SettingSaver plugin
					for index,key in ipairs(tmpKey) do
						retVal[index]=key
					end
				else
					-- return the value
					retVal=string.decode(tmpVal[0])
				end
				
			else
				if tmpVal[1]~=nil then
					retVal={}
					for k,v in ipairs(tmpVal) do
						local tmpReturnedKey=string.decode(tmpKey[k])
						retVal[tmpReturnedKey]=string.decode(v)
						if Turbine.Shell.IsCommand("0SST") and type(retVal[tmpReturnedKey])=="table" then
							local tmpRequest={}
							tmpRequest.Mode="Load"
							if Turbine.Shell.IsCommand("0SSS_Account") then
								tmpRequest.Scope="Account"
							elseif Turbine.Shell.IsCommand("0SSS_Server") then
								tmpRequest.Scope="Server"
							else
								tmpRequest.Scope="Character"
							end
							tmpRequest.Recursive=true
							tmpRequest.Debug=Turbine.Shell.IsCommand("0SSD")
							tmpRequest.File=settingsaverwindow.CurrentFile
							tmpRequest.Data=nil
							tmpRequest.Key=nil
							-- concatenate CurrentBranch, CurrentKey, and tmpReturnedKey to get new branch path
							
							if settingsaverwindow.CurrentKey~=nil and settingsaverwindow.CurrentKey~="" then
								if settingsaverwindow.CurrentBranch==nil or settingsaverwindow.CurrentBranch=="" then
									settingsaverwindow.CurrentBranch=settingsaverwindow.CurrentKey
								else
									settingsaverwindow.CurrentBranch=settingsaverwindow.CurrentBranch.."_"..settingsaverwindow.CurrentKey
								end
							end
							if settingsaverwindow.CurrentBranch==nil or settingsaverwindow.CurrentBranch=="" then
								if type(tmpReturnedKey)=="number" then
									tmpRequest.Branch="#"..tostring(tmpReturnedKey)
								else
									tmpRequest.Branch=tmpReturnedKey
								end
							else
								if type(tmpReturnedKey)=="number" then
									tmpRequest.Branch=settingsaverwindow.CurrentBranch.."_#"..tostring(tmpReturnedKey)
								else
									tmpRequest.Branch=settingsaverwindow.CurrentBranch.."_"..tmpReturnedKey
								end
							end
--Turbine.Shell.WriteLine("Setting Callback tmpReturnedKey:"..tostring(tmpReturnedKey).." retVal:"..tostring(retVal[tmpReturnedKey]))
							tmpRequest.Callback=function(val) retVal[tmpReturnedKey]=val end
							table.insert(settingsaverwindow.RequestQueue,1,tmpRequest) -- add to HEAD of queue
						end
					end
				else
					-- check for a set of keynames
					if tmpKey[1]~=nil then
						retVal={}
						for k,v in ipairs(tmpKey) do
							retVal[k]=string.decode(v)
						end
					end
				end
			end
			-- unload SettingSaver plugin
			Turbine.PluginManager.UnloadScriptState("SettingSaver");
			-- remove chat commands
			for k,v in pairs(settingsaverwindow.Cmds) do
				Turbine.Shell.RemoveCommand(v)
			end
			-- return requested value
			if settingsaverwindow.Callback~=nil then
				settingsaverwindow.Callback(retVal)
				settingsaverwindow.Callback=nil
			end
			if #settingsaverwindow.RequestQueue==0 then
				settingsaverwindow.State="idle"
				settingsaverwindow:SetWantsUpdates(false)
			else
				settingsaverwindow.State="init"
			end
		end
	end
end
settingsaverwindow.State="idle"
SaveSetting=function(scope,file,branch,key,data,recursive,debug)
	if branch==nil then branch="" end
	if recursive==nil then recursive=false end
	if debug==nil then debug=false end
	local tmpRequest={}
	tmpRequest.Mode="Save"
	tmpRequest.Scope=scope
	tmpRequest.Recursive=recursive
	tmpRequest.Debug=debug
	tmpRequest.File=file
	if type(data)=="table" then
		tmpRequest.Data={}
		tmpRequest.Key={}
		local index=1
		for k,v in pairs(data) do
			if type(v)~="table" then
				tmpRequest.Data[index]=v
				tmpRequest.Key[index]=k
				index=index+1
			end
		end
	else
		tmpRequest.Data=data
		tmpRequest.Key=key
	end
	tmpRequest.Branch=branch
	tmpRequest.Callback=nil
	table.insert(settingsaverwindow.RequestQueue,tmpRequest) -- add to queue
	if recursive and type(data)=="table" then
		for k,v in pairs(data) do
			if type(v)=="table" then
				-- recursively call SaveSetting to save the subtable
				SaveSetting(scope,file,branch.."_"..string.encode(k),nil,v,true,debug)
			end
		end
	end
	if settingsaverwindow.State=="idle" then settingsaverwindow.State="init" end
	settingsaverwindow:SetWantsUpdates(true)
end
LoadSetting=function(scope,file,branch,key,callback,recursive,debug)
	if branch==nil then branch="" end
	if recursive==nil then recursive=false end
	if debug==nil then debug=false end
	-- what exactly does recursive do in this context?
	local tmpRequest={}
	tmpRequest.Scope=scope
	tmpRequest.Mode="Load"
	tmpRequest.Recursive=recursive
	tmpRequest.Debug=debug
	tmpRequest.File=file
	tmpRequest.Data=data
	tmpRequest.Branch=branch
	tmpRequest.Key=key
	tmpRequest.Callback=callback
	table.insert(settingsaverwindow.RequestQueue,tmpRequest) -- add to queue
	if settingsaverwindow.State=="idle" then settingsaverwindow.State="init" end
	settingsaverwindow:SetWantsUpdates(true)
end
SettingQueueCallback=function(callback)
	-- this just queues a callback to be executed after all other queued commands are completed
	-- this allows plugins to be sure their data is all read before being processed
	local tmpRequest={}
	tmpRequest.Mode="Delay"
	tmpRequest.Callback=callback
	table.insert(settingsaverwindow.RequestQueue,tmpRequest)
	if settingsaverwindow.State=="idle" then settingsaverwindow.State="init" end
	settingsaverwindow:SetWantsUpdates(true)
end
SettingQueueCommand=function(scope,file,command,callback,debug)
	local tmpRequest={}
	tmpRequest.Scope=scope
	tmpRequest.Mode="Exec"
	tmpRequest.Debug=debug
	tmpRequest.File=file
	tmpRequest.Callback=callback
	tmpRequest.Command=command
	table.insert(settingsaverwindow.RequestQueue,tmpRequest)
	if settingsaverwindow.State=="idle" then settingsaverwindow.State="init" end
	settingsaverwindow:SetWantsUpdates(true)
end